/************************************************************************ *********
 * Proiectarea cu Microprocesoare
 * Laboratorul 1
 * Digital I/O
 *********************************************************************************/

#include <avr\io.h>
#include <avr\interrupt.h>

/************************************************************************ 
 * DEFINE-uri pentru parametrizarea modulului LCD
 ************************************************************************/
#define LcdDATA_DDR		DDRA			// Portul pe care conectam firele de date la LCD-ul
#define LcdDATA_PORT		PORTA
#define LcdDATA_PIN		PINA

#define LcdCMD_DDR		DDRA			// Portul pe care conectam firele de comenzi la LCD
#define LcdCMD_PORT		PORTA
#define LcdCMD_PIN		PINA

#define LcdD4			 PA0				// Pin-ul pentru firul de date D4 de pe LCD
#define LcdD5			 PA1				// Pin-ul pentru firul de date D5 de pe LCD
#define LcdD6			 PA2				// Pin-ul pentru firul de date D6 de pe LCD
#define LcdD7			 PA3				// Pin-ul pentru firul de date D7 de pe LCD

#define LcdRS			 PA4			    // Pinul pentru selectare operatie (LCD)
#define LcdRW			 PA5				// Pinul pentru Read/ Write (LCD)
#define LcdE			 PA6				// Pinul de Enable (LCD)

#define	SIZE	3
#define BUTTON_DDR		DDRC
#define	BUTTON_PORT		PORTC
#define	BUTTON_PIN		PINC
#define	OK				PC0
#define UP				PC1
#define LEFT			PC2
#define RIGHT			PC3
#define DOWN			PC4

#define UP_KEY		0
#define DOWN_KEY	1
#define	LEFT_KEY	2
#define	RIGHT_KEY	3
#define OK_KEY		5

#define BLANK	'_'
#define DELAY	40000

#define LCD_INSTR_4wire 		0x28 	// 4 fire date, font 5x8
#define LCD_INSTR_display 		0x0C 	// Display On, Cursor On, Blinking On ( 1 Display Cursor Blink )
#define LCD_INSTR_clearDisplay  	0x01 	// Clear Display
#define LCD_INSTR_returnHome		0x02 	// Return Cursor and LCD to Home Position
#define LCD_INSTR_nextLine 		0xC0 	// Return Cursor and LCD to Home Position
#define LCD_INSTR_gotoCGRAM		0x40	// go to Character Generator RAM


#define RS_HIGH()				LcdCMD_PORT |= _BV(LcdRS)
#define RS_LOW()				LcdCMD_PORT &= ~_BV(LcdRS)
 
#define RW_HIGH()				LcdCMD_PORT |= _BV(LcdRW)
#define RW_LOW()				LcdCMD_PORT &= ~_BV(LcdRW)
 
#define ENABLE()				LcdCMD_PORT |= _BV(LcdE)
#define DISABLE()				LcdCMD_PORT &= ~_BV(LcdE)

static char outputBoard[SIZE][SIZE], privateBoard[SIZE][SIZE];
char lastKeyPressed;

int delay, x, y;

void init();
void makeAMove();
int eval();
int getKeyPressed();

/*************
 * API LCD   *
 *************/
void LCD_init();                // Initializare modul LCD.Trebuie apelata inainte de a se face orice operatie cu LCD-ul
void LCD_writeInstruction(unsigned char _instruction);	// Trimite o instructiune catre lcd (vezi datasheet)
void LCD_writeData(unsigned char _data);		// Trimite date catre LCD pentru afisare
void LCD_write(unsigned char _byte);			// trimite un byte catre LCD la modul general (nu conteaza daca e instructiune sau date)
void LCD_waitNotBusy();					// Asteptam pana cand lcd-ul devine disponibil pt o noua comanda
void LCD_waitInstructions(unsigned char _instructions);	// Asteapta un numar de cicli de ceas.
void LCD_print(char* _msg);				// Afiseaza imformatia pe LCD (doar 1 linie, primele 16 caractere)
void LCD_print2(char* _msg1, char* _msg2);		// Afisare pe 2 lini pe LCD
void LCD_printDecimal2u(unsigned int _n);		// Afisare numar in baza 10 pe LCD
void LCD_printHexa(unsigned int _n);			// Afisare numar in baza 16 pe LCD

void LCD_printBoard();

int main()
{
	init();
	LCD_waitInstructions(6);
	LCD_printBoard();
	while(1)
	{
		getKeyPressed();
	}
	return 0;
}

/*****************************
 * IMPLEMENTARE API LCD.     *
 *****************************/
void LCD_init()
{
	LcdDATA_DDR |=  (1<<LcdD4)|(1<<LcdD5)|(1<<LcdD6)|(1<<LcdD7);	// setam pinii de date ca pini de iesire
	LcdCMD_DDR  |=  (1<<LcdRS)|(1<<LcdRW)|(1<<LcdE);				// setam pinii de comenzi ca pini de iesire

	LCD_waitNotBusy();
	
	LcdCMD_PORT   &= ~(1<<LcdRS);									// Setam linia RS pe low
	LcdCMD_PORT   &= ~(1<<LcdRW);									// Setam linia RW pe low (acum suntem in modul de trimis instructiuni)
	LcdDATA_PORT  &= ~(1<<LcdD4)&~(1<<LcdD6)&~(1<<LcdD7); 		// Specificam ca vrem 4 fire de date, prima comanda (LcdD5 activ, restul nu)
	LcdDATA_PORT  |=  (1<<LcdD5);									// setam pinii de comenzi ca pini de iesire
	
	LcdCMD_PORT |=  (1<<LcdE);					// Setam linia E(nable) pe high; aceasta ii specifica LCD-ului sa preia datele
	LCD_waitInstructions(6);					// Asteptam o perioada de timp T
	LcdCMD_PORT &= ~(1<<LcdE );				// Setam linia E(nable) pe low; transferul s-a terminat

	LCD_writeInstruction(LCD_INSTR_4wire); 		// Incarcam comanda: 4 bit data, 2 lines, 5x8 font
	LCD_writeInstruction(LCD_INSTR_display); 	// Display On, Cursor On, Blinking On
	LCD_writeInstruction(0x06);					// Increment, no shift
	LCD_writeInstruction(0x01);					// Clear Display

	
}

void LCD_writeInstruction(unsigned char _instruction)
{
	/* TODO */					
	// asteptam ca LCD-ul sa fie liber sa primeasca comenzile
	LCD_waitNotBusy();	
	// setam pinul RS pe low (low=instructiuni, high=date)
	RS_LOW();
	// setam pinul RW pe low (suntem in modul de comenzi acum)
	RW_LOW();
	// apelam procedura ce trimite byte-ul pe firele de date
	LCD_write(_instruction);
}


void LCD_writeData(unsigned char _data)
{
        /* TODO */
	// asteptam ca LCD-ul sa fie liber sa primeasca comenzile
	LCD_waitNotBusy();	
	// setam pinul RS pe high (low=instructiuni, high=date)
	RS_HIGH();
	// setam pinul RW pe low (suntem in modul de comenzi acum)
	RW_LOW();
	// apelam procedura ce trimite byte-ul pe firele de date
	LCD_write(_data);

}

void LCD_write(unsigned char _byte)
{
        /* TODO */
	PORTA &= 0xF0;
	
	// scriem cei mai semnificativi 4 biti pe firele de date

	PORTA |= (_byte >> 4);
	
	/*PORTA |= _byte & (1<<PA0);
	PORTA |= _byte & (1<<PA1);
	PORTA |= _byte & (1<<PA2);
	PORTA |= _byte & (1<<PA3);*/

	// Setam Pinul E pe high
	ENABLE();
	// Asteptam o perioada de timp T
	LCD_waitInstructions(6);
	// Setam Pinul E pe low
	DISABLE();
	// Asteptam o perioada de timp T
	LCD_waitInstructions(6);
	
	PORTA &= 0xF0;
	// scriem cei mai putin semnificativi 4 biti pe firele de date
			
	PORTA |= (_byte&0x0F);
	
	// Setam Pinul E pe high
	ENABLE();
	// Asteptam o perioada de timp T
	LCD_waitInstructions(6);
	// Setam Pinul E pe low
	DISABLE();
}

void LCD_waitNotBusy()
{
	unsigned char _loop = 1;

	while (_loop)
	{
		LcdDATA_DDR &= ~(1<<LcdD4 | 1<<LcdD5 | 1<<LcdD6 | 1<<LcdD7);	// Setam pinii de date de la LCD pe in pt a citi busy flag
		LcdDATA_PORT &= ~(1<<LcdD4 | 1<<LcdD5 | 1<<LcdD6 | 1<<LcdD7); 	// Dezactivam pullup resistor pentru pinii de in
		
		LcdCMD_PORT &= ~(1<<LcdE);					// Setam pin-ul e pe low; ar trebui sa fie deja low, doar ne asiguram
		LcdCMD_PORT &= ~(1<<LcdRS);					// Setam pinul RS pe low
		LcdCMD_PORT |=  (1<<LcdRW);					// Setam pinul RW pe high (acum suntem in modul de interogare busy/adr)
		
		LcdCMD_PORT |= (1<<LcdE);						// Setam Pinul E pe high
		LCD_waitInstructions(6);						// Asteptam o perioada de timp T
		_loop = LcdDATA_PIN & (1<<LcdD7);				// Citim busy flag-ul
		LcdCMD_PORT &= ~(1<<LcdE);						// Setam Pinul E pe low
		
		LCD_waitInstructions(6);						// Asteptam o perioada de timp T
		
		LcdCMD_PORT |= (1<<LcdE);						// Setam Pinul E pe high
		LCD_waitInstructions(6);						// Asteptam o perioada de timp T
		LcdCMD_PORT &= ~(1<<LcdE);						// Setam Pinul E pe low

		LcdDATA_DDR |= (1<<LcdD4 | 1<<LcdD5 | 1<<LcdD6 | 1<<LcdD7); // Setam Portul de LCD ca port de iesire la loc
	}
}

void LCD_waitInstructions(unsigned char _instructions)
{
	while (_instructions--)
		;
}


void LCD_printDecimal2u(unsigned int _n)
{
	unsigned char tmp=0;
	
	// Extragem sutele
	while(_n>=100)
		_n-=100;

	while(_n>=10){
		tmp++;
		_n-=10;
	}

	LCD_writeData(tmp+'0');
	LCD_writeData(_n+'0');
}

void LCD_printHexa(unsigned int _n)
{
	unsigned char _tmp = _n>>4;
	if (_tmp>9)
		_tmp += 'A'-10;
	else
		_tmp += '0';
	LCD_writeData( _tmp );
	_tmp = _n & 0x0F;
	if (_tmp>9)
		_tmp += 'A'-10;
	else
		_tmp += '0';
	LCD_writeData( _tmp );
}


void LCD_print(char* _msg)
{
	unsigned char i=0;
	for( ; _msg[i]!=0 && i<16; i++)
		LCD_writeData( _msg[i] );
}

void LCD_print2(char* _msg1, char* _msg2)
{
	LCD_writeInstruction(LCD_INSTR_clearDisplay);
	LCD_print(_msg1);
	LCD_writeInstruction(LCD_INSTR_nextLine);
	LCD_print(_msg2);
}

void LCD_printBoard()
{
	LCD_writeInstruction(LCD_INSTR_clearDisplay);
	unsigned char i = 0;
	outputBoard[x][y] = '-';
	for(i = 0; i < 9; ++i)
		LCD_writeData( outputBoard[ i/3][ i%3 ] );
	LCD_writeInstruction(LCD_INSTR_nextLine);
	if(x == 0)
		LCD_writeData( '0' );
	else if(x == 1)
		LCD_writeData( '1' );
	else if(x == 2)
		LCD_writeData( '2' );
	
	LCD_writeData(' ');
	
	if(y == 0)
		LCD_writeData( '0' );
	else if(y == 1)
		LCD_writeData( '1' );
	else if(y == 2)
		LCD_writeData( '2' );
}

void init() 
{
	unsigned char i, j;
	
	BUTTON_PORT = 0xff;
	BUTTON_DDR = 0;
	
	LCD_init();
	
	x = 0, y = 0;
	delay = DELAY, lastKeyPressed = -1;
	for(i = 0; i < 3; ++i)	
		for(j = 0; j < 3; ++j)
		{
			privateBoard[i][j] = 0;
			outputBoard[i][j] = BLANK; 
		}
}

void makeAMove()
{
	
}

/* Intoarce 0 daca jocul nu e finalizat, 1 daca e castigat de 'X', 2 daca e castigat de '0'*/
int eval()
{
	unsigned char i, j;
	
	return 0;
}

int getKeyPressed()
{
	if(delay > 0)
		delay--;

	if(~BUTTON_PIN & (1 << UP))
	{
		if(lastKeyPressed == UP_KEY && delay > 0)
			return 0;
		outputBoard[x][y] = BLANK;
		/* Verificare limite. */
		if(y == 0)
			y = 2;
		else
			y--;
		lastKeyPressed = UP_KEY;
		delay = DELAY;
		LCD_printBoard();
	}
	else if(~BUTTON_PIN & (1 << LEFT))
	{
		if(lastKeyPressed == LEFT_KEY && delay > 0)
			return 0;
		
		outputBoard[x][y] = BLANK;
		/* Verificare limite. */
		if(x == 0)
			x = 2;
		else
			x--;
		lastKeyPressed = LEFT_KEY;
		delay = DELAY;
		LCD_printBoard();
	}
	else if(~BUTTON_PIN & (1 << RIGHT))
	{
		if(lastKeyPressed == RIGHT_KEY && delay > 0)
			return 0;
		outputBoard[x][y] = BLANK;
		/* Verificare limite. */
		if(x == 2)
			x = 0;
		else
			x++;
		lastKeyPressed = RIGHT_KEY;
		delay = DELAY;
		LCD_printBoard();
	}
	else if(~BUTTON_PIN & (1 << DOWN))
	{
		if(lastKeyPressed == DOWN_KEY && delay > 0)
			return 0;
		outputBoard[x][y] = BLANK;
		/* Verificare limite. */
		if(y == 2)
			y = 0;
		else
			y++;
		lastKeyPressed = DOWN_KEY;
		delay = DELAY;
		LCD_printBoard();
	}
	else if(~BUTTON_PIN & (1 << OK))
	{
		if(lastKeyPressed == OK_KEY && delay > 0)
			return 0;
		lastKeyPressed = OK_KEY;
		if(outputBoard[x][y] == BLANK)
			
		delay = DELAY;
		LCD_printBoard();
	}
	
	return 0;
}